# Copyright (c) James Yates Farrimond. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# Modification, are permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY JAMES YATES FARRIMOND ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JAMES YATES FARRIMOND OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation
# are those of the authors and should not be interpreted as representing
# official policies, either expressed or implied, of James Yates
# Farrimond.

'''
Frycook depends on recipes and cookbooks to define how systems are
built.
'''

from cookbook_template import Cookbook  # noqa
from recipe_template import Recipe, RecipeException  # noqa

__version__ = '0.3.9'



# Copyright (c) James Yates Farrimond. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# Modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY JAMES YATES FARRIMOND ''AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL JAMES YATES FARRIMOND OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of James Yates Farrimond.

'''
Recipes define subsystems that are distinct parts of larger systems.
They are the basic units of setup in frycook.  Generally a recipe
corresponds to an os-level package that needs to be installed or
configured.
'''
import os
import os.path
import re
import shutil
import stat

import cuisine
from fabric.api import local
from mako.lookup import TemplateLookup


class RecipeException(Exception):
    '''
    A RecipeException exception is raised for exceptional conditions
    encountered while using the Recipe class.
    '''
    pass


class FileMetaDataTracker(object):
    '''
    A FileMetaDataTracker object keeps track of owner, group, and
    permissions for files and directories during a push_package_fileset
    operation.  This hinges on a file named fck_metadata.txt being
    encountered in the directory being examined.  This file should have
    each line contain the text <filename>:<owner>:<group>:<perms>, where
    <filename> is either a file name or '.' for the directory itself,
    <owner> is the owner's account name, <group> is the group's name,
    and <perms> is the permissions string..
    '''

    tagfile = 'fck_metadata.txt'

    def __init__(self):
        '''
        Start with a blank metadata dictionary.
        '''
        self.metadata = {}

    def check_directory(self, root, dirs, files):
        '''
        Read in the metadata for the given directory from the
        fck_metadata.txt file in the directory, or remember metadata
        from previous calls if no fck_metadata.txt file is encountered.

        The input parameters are expected to be the values returned from
        a call to os.walk()

        :type root: string
        :param root: directory being examined
        :type dirs: list of strings
        :param dirs: list of sub-directories under the root directory
        :type files: list of strings
        :param files: list of the files in the root directory
        '''
        if root not in self.metadata:
            self.metadata[root] = (None, None, None, )

        if self.tagfile in files:
            for line in open(os.path.join(root, self.tagfile)):
                parts = line.split(':')
                path = parts[0].strip()
                if len(path) > 0 and path[0] != '#':
                    owner = parts[1].strip()
                    group = parts[2].strip()
                    if len(parts) > 3:
                        perms = parts[3].strip()
                    else:
                        perms = None

                    if path == '.':
                        self.metadata[root] = (owner, group, perms, )
                        for dn in dirs:
                            self.metadata[os.path.join(root, dn)] = (
                                owner, group, perms, )
                    else:
                        self.metadata[os.path.join(root, path)] = (
                            owner, group, perms, )

    def get_metadata(self, path, filename=''):
        '''
        Lookup the owner, group, and permissions for the given path and
        filename from the metadata dictionary in this object.  If no
        filename is specified, the data is retrieved for the directory
        specified in path.

        :type path: string
        :param path: path to file to get metadata for
        :type filename: string
        :param filename: name of file to get metadata for (leave blank if just getting directory metadata)

        :rtype: tuple of strings
        :return: tuple containing (<owner>, <group>, <perms>, )
        '''
        fq = os.path.join(path, filename)
        if fq in self.metadata:
            return self.metadata[fq]
        elif path in self.metadata:
            return self.metadata[path]
        else:
            return None, None, None


class FileDeleter(object):
    '''
    A FileDeleter object deletes unwanted files from a directory on a
    remote server.  This hinges on a file named fck_delete.txt being
    encountered in the directory being examined.  This file should have
    each line contain the name of a file to be deleted.
    '''

    tagfile = 'fck_delete.txt'

    def check_directory(self, root, files, remote_rootpath):
        '''
        Examine the given directory, check for a fck_delete.txt file in
        the directory, and if it exists delete all remote files named in
        it.

        :type root: string
        :param root: local directory possibly containing fck_delete.txt
        :type files: list of strings
        :param files: list of the files in the local root directory
        :type remote_rootpath: string
        :param remote_rootpath: path on remote server to delete files from
        '''

        if self.tagfile in files:
            for line in open(os.path.join(root, self.tagfile)):
                delfile = os.path.join(remote_rootpath, line.strip())
                cuisine.file_unlink(delfile)


class Recipe(object):
    '''
    The Recipe class is the base class for all recipes to subclass.  It
    defines the framework for recipes and some helper functions.  By
    itself it doesn't do much.

    It has a set of helper functions that are hooks called by frycooker
    to apply the recipe to a remote server::

        handle_pre_apply_message()
        handle_post_apply_message()
        run_apply()
        run_messages()

    It has another set of helper functions used within recipes for
    copying files to remote servers::

        get_local_file_perms()
        push_file()
        push_template()
        push_package_file_set()

    It has a final set of helper functions used within recipes for
    managing git repos on remote servers::

        push_git_repo()
        clone_git_repo()
        update_git_repo()
        is_git_repo()
        ensure_git_repo()
    '''

    def __init__(self, settings, environment, ok_to_be_rude, no_prompt):
        '''
        Initialize the recipe object with the settings and environment
        dictionaries.

        :param dict sttings: settings dictionary
        :param dict environment: metadata dictionary
        :param boolean ok_to_be_rude: is it ok to interrupt your users?
        :param boolean no_prompt: should we prompt the user?
        '''
        self.settings = settings
        self.environment = environment
        self.ok_to_be_rude = ok_to_be_rude
        self.no_prompt = no_prompt
        self.mylookup = TemplateLookup(
            directories=[self.settings["package_dir"]])

    #######################
    ######## APPLY ########
    #######################

    pre_apply_message = ""

    def handle_pre_apply_message(self):
        '''
        Print the pre-apply message for the user and wait for him/her to hit
        return before continuing.
        '''
        if self.pre_apply_message:
            header = "pre-apply message from %s:" % self.__class__.__name__
            print '=' * len(header)
            print header
            print '=' * len(header)
            print self.pre_apply_message
            if not self.no_prompt:
                raw_input('press enter to continue')

    post_apply_message = ""

    def handle_post_apply_message(self):
        '''
        Print the post-apply message for the user and wait for him/her to hit
        return before continuing.
        '''
        if self.post_apply_message:
            header = "post-apply message from %s:" % self.__class__.__name__
            print '=' * len(header)
            print header
            print '=' * len(header)
            print self.post_apply_message
            if not self.no_prompt:
                raw_input('press enter to continue')

    def pre_apply_checks(self, computer):
        '''
        Define checks to run before applying the recipe.  The base class checks
        that the designated computer exists in the environment dictionary.
        This is a good place to check other things in your environment
        dictionary that the apply function expects to be there.  Override this
        function in your subclass of Recipe if you have any checks to perform.
        Raise a RecipeException if you encounter exceptional conditions.  Be
        sure to call the base class function to make sure its checks get done
        as well.

        :param string computer: name of computer to apply recipe checks to
        :raises RecipeException: raised if the computer name is not in the environment being processed
        '''
        # make sure the computer is in our enviro
        if computer not in self.environment["computers"]:
            raise RecipeException(
                "computer %s not defined in environment" % computer)

    def apply(self, computer):
        '''
        Define the actions to take place when this recipe is applied to a
        computer.  Override this function in your subclass of Recipe if you
        have any actions to apply.  If you don't have any actions, then why are
        you creating a recipe?

        :type computer: string
        :param computer: name of computer to apply recipe to
        '''
        pass

    def run_apply(self, computer):
        '''
        Run the apply sequence of functions.  This is typically called by
        frycooker.

        Sequence::

          pre_apply_checks() -> apply()

        :type computer: string
        :param computer: name of computer to apply recipe to
        '''
        self.pre_apply_checks(computer)
        self.apply(computer)

    def run_messages(self):
        '''
        Print the pre and post apply messages only, as if the apply had been
        run.
        '''
        self.handle_pre_apply_message()
        self.handle_post_apply_message()

    ###############################
    ######## FILE HANDLING ########
    ###############################

    def get_local_file_perms(self, local_name):
        '''
        Get a string of the permissions set on a local file.

        :type local_name: string
        :param local_name: path to file on local file system

        :rtype: string
        :return: string containing perms of file, ie. '655'
        '''
        bit_mode = stat.S_IMODE(os.stat(local_name).st_mode)
        user_perms = (bit_mode & stat.S_IRWXU) >> 6
        group_perms = (bit_mode & stat.S_IRWXG) >> 3
        other_perms = bit_mode & stat.S_IRWXO
        string_perms = "%s%s%s" % (user_perms, group_perms, other_perms)
        return string_perms

    def push_file(self, local_name, remote_name, owner, group, perms=None):
        '''
        Copy a file to a remote server if the file is different or doesn't
        exist.

        :type local_name: string
        :param local_name: path within packages dir of file to upload (path + filename)
        :type remote_name: string
        :param remote_name: remote path to write file to (path + filename)
        :type owner: string
        :param owner: owner of the file
        :type group: string
        :param group: group of the file
        :type perms: string
        :param perms: permissions for the file, ie. '655'
        '''
        local_name = os.path.join(self.settings["package_dir"], local_name)
        cuisine.file_upload(remote_name, local_name)
        if not perms:
            perms = self.get_local_file_perms(local_name)
        cuisine.file_attribs(
            remote_name, mode=perms, owner=owner, group=group)

    def push_template(self, templatename, out_path, enviro,
                      owner, group, perms=None):
        '''
        Process a template file and push its contents to a remote server if
        it's different than what's already there.

        :type templatename: string
        :param templatename: path within packages dir of template file to process (path + filename)
        :type out_path: string
        :param out_path: path on remote server to write file to (path + filename)
        :type enviro: dict
        :param enviro: environment dictionary for template engine
        :type owner: string
        :param owner: owner of the templated file
        :type group: string
        :param group: group of the templated file
        :type perms: string
        :param perms: permissions for the templated file, ie. '655'
        '''
        mytemplate = self.mylookup.get_template(templatename)
        try:
            buff = mytemplate.render(**enviro)
        except Exception, e:
            raise RecipeException(
                "Error rendering template %s: %s" % (templatename, e))
        cuisine.file_write(out_path, buff, check=True)
        local_name = os.path.join(self.settings["package_dir"], templatename)
        if not perms:
            perms = self.get_local_file_perms(local_name)
        cuisine.file_attribs(
            out_path, mode=perms, owner=owner, group=group)

    def _push_package_file_set(self, package_name, template_env):
        '''
        Implement the file copying and deleting portion of the
        push_package_file_set operation.  The calling function sets up the
        template environment, then calls this one.

        :type package_name: string
        :param package_name: name of package to process, corresponds to directory in packages directory
        :type template_env: dict
        :param template_env: environment dictionary for template engine
        '''
        metadata = FileMetaDataTracker()
        deleter = FileDeleter()
        work_dir = os.path.join(self.settings["package_dir"], package_name)
        os.chdir(work_dir)
        for root, dirs, files in os.walk('.'):
            metadata.check_directory(root, dirs, files)
            owner, group, perms = metadata.get_metadata(root)
            remote_root = '/'+root.lstrip('.')
            cuisine.dir_ensure(
                remote_root, owner=owner, group=group, mode=perms)
            for filename in files:
                fq_filename = os.path.join(remote_root, filename).lstrip('/')
                if (re.search(self.settings["file_ignores"], filename) is None
                        and filename != metadata.tagfile
                        and filename != deleter.tagfile):
                    owner, group, perms = metadata.get_metadata(root, filename)
                    base_path, ext = os.path.splitext(fq_filename)
                    if ext == '.tmplt':
                        self.push_template(os.path.join(package_name,
                                                        fq_filename),
                                           os.path.join('/', base_path),
                                           template_env, owner, group, perms)
                    else:
                        self.push_file(os.path.join(work_dir, fq_filename),
                                       os.path.join('/', fq_filename),
                                       owner, group, perms)
            deleter.check_directory(root, files, remote_root)

    def push_package_file_set(self, package_name, computer_name, aux_env=None):
        '''
        Copy a set of files to a remote server, maintaining the same directory
        structure and processing any templates encountered. This copies the
        files to the root of the destination, so that things like /etc/hosts or
        /etc/nginx/nginx.conf get put in the right place.

        First create a directory with the same name as package_name in your
        root packages directory.  Then create a directory structure under that
        that mirrors the target machine and all files will get copied there in
        the correct place.

        For template file processing, a default environment dictionary will be
        passed in consisting of::

          {"computer": host_env["computers"][computer_name]}

        Template files have a .tmplt extension.

        If aux_env is given, it will be added to the default dictionary using
        dict.update() after the default dictionary is constructed.  This means
        that if you pass in an aux_env dictionary with a key called "computer",
        that that key will over-write the default key of that name.

        If a file named fck_metadata.txt is encountered in a directory, then
        it's expected to have contents of lines consisting of
        <path>:<owner>:<group>:<perms> that specifies the owner, group, and
        permissions for the path specified (directory-relative).

        If a file named fck_delete.txt is encountered in a directory, then it's
        expected to have contents of lines containing names of files to delete,
        one per line.  This way you can clean out a directory as well as copy
        files to it.

        :type package_name: string
        :param package_name: name of package to process, corresponds to directory in packages directory
        :type template_env: dict
        :param template_env: environment dictionary for template engine
        :type aux_env: dict
        :param aux_env: additional key/value pairs for the template environment
        '''
        template_env = {"computer":
                        self.environment["computers"][computer_name]}
        if aux_env is not None:
            template_env.update(aux_env)
        self._push_package_file_set(package_name, template_env)

    def append_line_to_file(self, tag, add_line, filepath):
        '''
        Append a line to a file on the remote filesystem if it's not
        there already.  Look for the tag to see if the line is there
        already, in case the existing line has different spacing or
        tabbing than the new line.

        :type tag: string
        :param tag: tag to look for in existing lines
        :type add_line: string
        :param add_line: line to append to file
        :type filepath: string
        :param filepath: fully-qualified path to remote file
        '''
        old_contents = cuisine.file_read(filepath)
        eol = cuisine.text_detect_eol(old_contents)
        old_contents = old_contents.rstrip(eol)
        old_contents = old_contents.split(eol)
        has_line = False
        for line in old_contents:
            print line
            if line.find(tag) != -1:
                has_line = True
                continue
        if not has_line:
            old_contents.append(add_line)
            cuisine.file_write(filepath, eol.join(old_contents) + eol)

    def find_replace_in_file(self, old_text, new_text, filepath):
        '''
        Find and replace text in a file on the remote filesystem.

        :type old_text: string
        :param tag: text to replace
        :type new_text: string
        :param add_line: text to replace with
        :type filepath: string
        :param filepath: fully-qualified path to remote file
        '''
        old_contents = cuisine.file_read(filepath)
        eol = cuisine.text_detect_eol(old_contents)
        old_contents = old_contents.split(eol)
        new_contents = []
        for line in old_contents:
            print line
            new_line = line.replace(old_text, new_text)
            new_contents.append(new_line)
        cuisine.file_write(filepath, eol.join(new_contents))

    ##############################
    ######## GIT HANDLING ########
    ##############################

    def push_git_repo(self, computer, user, group, git_url, target_path):
        '''
        Make a local clone of the repo in git_url into the temp directory
        specified in the settings file, then rsync it to the remote path.

        :type computer: string
        :param computer: computer name to push to
        :type user: string
        :param user: user to own the files in the repo
        :type group: string
        :param group: group to own the files in the repo
        :type git_url: string
        :param git_url: git url of repo (probably from github)
        :type target_path: string
        :param target_path: root path on remote server to copy git repo to
        '''
        rsync_command = ('rsync -qrlptz --delete --delete-excluded '
                         '--exclude=.svn --exclude=.git')
        tmp_path = os.path.join(self.settings["tmp_dir"],
                                'push_git_repo/repo/')
        if not os.path.exists(tmp_path):
            local('git clone %s %s' % (git_url, tmp_path))
        else:
            local('cd %s && git pull' % tmp_path)
        local('%s %s root@%s:%s' %
              (rsync_command, tmp_path, computer, target_path))
        cuisine.sudo('chown -R %s:%s %s' % (user, group, target_path))
        shutil.rmtree(tmp_path)

    def clone_git_repo(self, user, git_url, target_path):
        '''
        Clone a git repo on a remote server.

        :type git_url: string
        :param git_url: git url of repo (probably from github)
        :type target_path: string
        :param target_path: root path on remote server to clone git repo into
        '''
        cuisine.sudo('sudo -Hi -u %s git clone %s %s' %
                     (user, git_url, target_path))

    def update_git_repo(self, user, git_url, target_path):
        '''
        Update an existing git repo on a remote server.

        :type git_url: string
        :param git_url: git url of repo (probably from github)
        :type target_path: string
        :param target_path: root path on remote server to update git repo in
        '''
        cuisine.sudo('cd %s && sudo -u %s git pull' % (target_path, user))

    def is_git_repo(self, target_path):
        '''
        Make sure the target path exists on the remote computer and is
        really a git repo.

        :type git_url: string
        :param git_url: git url of repo (probably from github)
        :type target_path: string
        :param target_path: root path on remote server to check git repo

        :rtype: boolean
        :return: True if repo already existed, False if not

        '''
        if not cuisine.dir_exists(os.path.join(target_path, '.git')):
            return False
        ret = cuisine.sudo('cd %s && git status' % target_path)
        if not ret.succeeded:
            return False
        return True

    def ensure_git_repo(self, user, git_url, target_path):
        '''
        Make sure a git repo exists in the target path on the remote
        computer.

        :type git_url: string
        :param git_url: git url of repo (probably from github)
        :type target_path: string
        :param target_path: root path on remote server to check git repo

        :rtype: boolean
        :return: True if repo already existed, False if not
        '''
        if not cuisine.dir_exists(target_path):
            self.clone_git_repo(user, git_url, target_path)
            return False
        return True



#!/env/python

# Copyright (c) James Yates Farrimond. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# Modification, are permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY JAMES YATES FARRIMOND ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JAMES YATES FARRIMOND OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation
# are those of the authors and should not be interpreted as representing
# official policies, either expressed or implied, of James Yates
# Farrimond.

'''
Frycooker is the program that takes all your carefully coded recipes and
cookbooks and applies them to computers.
'''

import argparse
import json
import os
import shutil
import sys
import tempfile

import cuisine
from fabric.api import env
from fabric.network import disconnect_all

import cookbooks
import recipes


def replace_tilde_in_path(path):
    return path.replace('~', os.environ['HOME'])


def get_args():
    '''
    Parse the command-line arguments, using python's argparse library.

    :rtype: argparse args object
    :return: object containing attributes for all possible command-line arguments
    '''
    recipe_names = recipes.recipes.keys()
    recipe_names.sort()
    cookbook_names = cookbooks.cookbooks.keys()
    cookbook_names.sort()

    parser = argparse.ArgumentParser(description='Setup machines.')
    parser.add_argument('-a', '--apply', action='store_true', default=False,
                        help='apply components to named servers')
    parser.add_argument('-c', '--cookbook', dest='cookbooks', action='append',
                        choices=cookbook_names,
                        help='cookbook to process (can specify multiple times)'
                        )
    parser.add_argument('-d', '--dryrun', action='store_true', default=False,
                        help='do not apply actions, just verify environment '
                        'and see which hosts to apply to')
    parser.add_argument('-e', '--environment', default='environment.json',
                        help='environment file')
    parser.add_argument('-k', '--keyfile',
                        help='full path to ssh key file to use')
    parser.add_argument('-m', '--messages', action='store_true', default=False,
                        help='do not apply actions, just print messages')
    parser.add_argument('-n', '--no-prompt', action='store_true', default=False,
                        dest='no_prompt', help='do not prompt user for '
                        'anything; good for automated scripts')
    parser.add_argument('-O', '--ok-to-be-rude', action='store_true',
                        default=False, dest='ok_to_be_rude',
                        help='ok to be rude to your users')
    parser.add_argument('-p', '--package-update', action='store_true',
                        default=False, dest='package_update',
                        help='update the package manager before '
                        'applying recipes/cookbooks')
    parser.add_argument('-P', '--param', dest='params', action='append',
                        help='extra parameters to pass in to recipes and '
                        'cookbooks (key:value) (can specify multiple times)')
    parser.add_argument('-r', '--recipe', dest='recipes', action='append',
                        choices=recipe_names,
                        help='recipe to process (can specify multiple times)')
    parser.add_argument('-s', '--settings', default='settings.json',
                        help='settings file')
    parser.add_argument('-S', '--sudo', action='store_true', default=False,
                        help='run all commands on client as sudo')
    parser.add_argument('-u', '--user', default='root',
                        help='user to ssh to host as')
    parser.add_argument('target', nargs='+',
                        help='computer or group to apply setup to')

    args = parser.parse_args()
    if args.keyfile is not None:
        args.keyfile = replace_tilde_in_path(args.keyfile)
    return args


def massage_enviro_paths(env):
    '''
    Recursively iterate through all the keys in the environment
    dictionary and look for any that contain the strings 'path' or 'dir'
    in them.  For the values corresponding to the keys that do, replace
    any instances of the '~' character with the running user's home
    directory path.

    :type env: dictionary
    :param env: environment dictionary
    '''
    for k, v in env.iteritems():
        if isinstance(v, dict):
            massage_enviro_paths(v)
        elif (isinstance(k, basestring) and
              isinstance(v, basestring) and
              (k.find('path') > -1 or k.find('dir') > -1)):
            env[k] = replace_tilde_in_path(v)


def load_settings(filename, params):
    '''
    Load the settings json file, massaging its environment paths in the
    process.

    :type filename: string
    :param filename: filename of settings file to read
    :type args: args object
    :param args: object containing attributes for all possible command-line parameters

    :rtype: dictionary
    :return: dictionary representation of settings
    '''
    settings = json.load(open(filename))
    settings["params"] = {}
    if params:
        for p in params:
            parts = p.split(':')
            if len(parts) == 2:
                settings["params"][parts[0]] = parts[1]
    massage_enviro_paths(settings)
    return settings


def load_enviro(filename):
    '''
    Load the environment json file, loading additionaly imported
    environment files and massaging environment paths in the
    process.

    :type filename: string
    :param filename: filename of environment file to read

    :rtype: dictionary
    :return: dictionary representation of environment
    '''
    try:
        enviro = json.load(open(filename))
    except Exception:
        print 'Error reading environment file %s' % filename
        raise
    massage_enviro_paths(enviro)

    for key in enviro:
        if 'imports' in enviro[key]:
            for imp_file in enviro[key]['imports']:
                imp_enviro = load_enviro(imp_file)
                for imp_key in imp_enviro:
                    enviro[key][imp_key] = imp_enviro[imp_key]
            del enviro[key]['imports']

    return enviro


class InvalidTarget(Exception):
    '''
    An InvalidTarget exception is raised for exceptional conditions
    encountered while generating target lists.
    '''
    pass


def generate_target_list(enviro, args):
    '''
    Get a list of computers to run against from the command-line
    arguments.  Translate group names into lists of computers.

    :type enviro: dictionary
    :param enviro: environment to read group lists from
    :type args: args object
    :param args: object containing attributes for all possible command-line parameters

    :rtype: list of strings
    :return: list of computer names to run against
    '''
    host_list = []
    for target in args.target:
        if target in enviro['computers']:
            host_list.append(target)
        elif target in enviro['groups']:
            host_list.extend(enviro['groups'][target]['computers'])
        else:
            raise InvalidTarget("Invalid target '%s' encountered" % target)
            sys.exit(2)
    return host_list


def generate_run_list(enviro, args):
    '''
    Get the lists of computers, recipes, cookbooks, and run list to run
    against.  A run list is a dictionary containing a list of recipes
    and cookbooks to run against each specified computer.  Typically the
    run list will just specify the same recipes and cookbooks for each
    computer, unless a list of computers was given and the apply (-a)
    command-line argument was specified.  In that case each computer
    could have different things run against them.

    :type enviro: dictionary
    :param enviro: environment to read group lists and computer components from
    :type args: args object
    :param args: object containing attributes for all possible command-line parameters

    :rtype: tuple of (dictionary, list, list, list)
    :return: (run list, host list, recipe list, cookbook list)
    '''
    host_list = generate_target_list(enviro, args)
    run_list = {}
    cookbooks = set()
    recipes = set()
    if args.apply:
        for host in host_list:
            run_list[host] = []
            for comp in enviro['computers'][host]["components"]:
                if comp["type"] == 'cookbook':
                    run_list[host].append(comp)
                    cookbooks.add(comp["name"])
                elif comp["type"] == 'recipe':
                    run_list[host].append(comp)
                    recipes.add(comp["name"])
    else:
        for host in host_list:
            run_list[host] = []
            if args.cookbooks:
                run_list[host].extend([{"type": "cookbook", "name": c} for c in args.cookbooks])
            if args.recipes:
                run_list[host].extend([{"type": "recipe", "name": r} for r in args.recipes])
        if args.cookbooks:
            cookbooks.update(args.cookbooks)
        if args.recipes:
            recipes.update(args.recipes)
    return run_list, host_list, recipes, cookbooks


def output_pre_apply_messages(recipe_list, cookbook_list, enviro, settings, args):
    '''
    Ouptput all the pre-apply messages for the specified recipes and
    cookbooks.

    :type recipe_list: list
    :param recipe_list: list of recipes to process
    :type cookbook_list: list
    :param cookbook_list: list of cookbooks to process
    :type enviro: dictionary
    :param enviro: environment dictionary
    :type settings: dictionary
    :param settings: settings dictionary
    :type args: args object
    :param args: object containing attributes for all possible command-line parameters
    '''
    for r in recipe_list:
        recipe = recipes.recipes[r](
            settings, enviro, args.ok_to_be_rude, args.no_prompt)
        recipe.handle_pre_apply_message()

    for c in cookbook_list:
        cookbook = cookbooks.cookbooks[c](
            settings, enviro, args.ok_to_be_rude, args.no_prompt)
        cookbook.handle_pre_apply_messages()


def output_post_apply_messages(recipe_list, cookbook_list, enviro, settings, args):
    '''
    Ouptput all the post-apply messages for the specified recipes and
    cookbooks.

    :type recipe_list: list
    :param recipe_list: list of recipes to process
    :type cookbook_list: list
    :param cookbook_list: list of cookbooks to process
    :type enviro: dictionary
    :param enviro: environment dictionary
    :type settings: dictionary
    :param settings: settings dictionary
    :type args: args object
    :param args: object containing attributes for all possible command-line parameters
    '''
    for r in recipe_list:
        recipe = recipes.recipes[r](
            settings, enviro, args.ok_to_be_rude, args.no_prompt)
        recipe.handle_post_apply_message()

    for c in cookbook_list:
        cookbook = cookbooks.cookbooks[c](
            settings, enviro, args.ok_to_be_rude, args.no_prompt)
        cookbook.handle_post_apply_messages()


def apply_recipes_cookbooks(enviro, settings, args, host_list, run_list):
    '''
    Apply all specified recipes and cookbooks to the requested hosts.

    :type enviro: dictionary
    :param enviro: environment dictionary
    :type settings: dictionary
    :param settings: settings dictionary
    :type args: args object
    :param args: object containing attributes for all possible command-line parameters
    :type host_list: list of strings
    :param host_list: list of hosts to run against
    :type run_list: dictionary
    :param run_list: dictionary of lists
    '''
    for host in host_list:
        env.host_string = host
        if args.user:
            env.user = args.user
        if args.keyfile:
            env.key_filename = args.keyfile

        if args.package_update:
            cuisine.package_update()

        try:
            for item in run_list[host]:
                if item["type"] == "recipe":
                    recipe = recipes.recipes[item["name"]](
                        settings, enviro, args.ok_to_be_rude, args.no_prompt)
                    recipe.run_apply(host)
                elif item["type"] == "cookbook":
                    cookbook = cookbooks.cookbooks[item["name"]](
                        settings, enviro, args.ok_to_be_rude, args.no_prompt)
                    cookbook.run_apply(host)
        finally:
            disconnect_all()


def main():
    '''
    Main function for the frycooker program.
    '''
    args = get_args()

    settings = load_settings(args.settings, args.params)
    enviro = load_enviro(args.environment)

    try:
        if args.sudo:
            cuisine.mode_sudo()

        tmp_dir = tempfile.mkdtemp(dir=settings["tmp_dir"])
        settings["tmp_dir"] = tmp_dir
        run_list, host_list, recipes, cookbooks = generate_run_list(enviro, args)

        if args.dryrun:
            print ("actions would be applied to the following hosts: %s" %
                   ', '.join(host_list))
            print ("environment to be used: %s" %
                   json.dumps(enviro, sort_keys=True, indent=4,
                              separators=(',', ': ')))
            sys.exit(0)

        output_pre_apply_messages(recipes, cookbooks, enviro, settings, args)
        if not args.messages:
            apply_recipes_cookbooks(enviro, settings, args, host_list, run_list)
        output_post_apply_messages(recipes, cookbooks, enviro, settings, args)

        shutil.rmtree(tmp_dir)

        print "actions completed successfully"
    except Exception, e:
        print "EXITING EARLY DUE TO AN EXCEPTION:"
        print e
        shutil.rmtree(tmp_dir)
        sys.exit(2)


if __name__ == "__main__":
    main()
